"""
Python SDK for reading/writing to/from pvsite database
"""

from .connection import DatabaseConnection
from .sqlmodels import (
    ClientSQL,
    DatetimeIntervalSQL,
    ForecastSQL,
    ForecastValueSQL,
    GenerationSQL,
    LatestForecastValueSQL,
    SiteSQL,
    StatusSQL,
)



""" Read pv generation functions """
import logging
import uuid
from datetime import datetime
from typing import List, Optional

from pvsite_datamodel.sqlmodels import ClientSQL, DatetimeIntervalSQL, GenerationSQL, SiteSQL
from sqlalchemy.orm import Session

from .utils import filter_query_by_datetime_interval

logger = logging.getLogger(__name__)


def get_pv_generation_by_client(
    session: Session,
    start_utc: Optional[datetime] = None,
    end_utc: Optional[datetime] = None,
    client_names: Optional[List[str]] = None,
) -> List[GenerationSQL]:
    """
    Get the generation data by client

    :param session: database session
    :param end_utc: search filters < on 'datetime_utc'. Can be None
    :param start_utc: search filters >= on 'datetime_utc'. Can be None
    :param client_names: optional list of provider names
    :return:list of pv yields
    """

    # start main query
    query = session.query(GenerationSQL)
    query = query.join(SiteSQL)
    query = query.join(ClientSQL)
    query = query.join(DatetimeIntervalSQL)

    # Filter by time interval
    query = filter_query_by_datetime_interval(query=query, start_utc=start_utc, end_utc=end_utc)

    if client_names is not None:
        query = query.filter(ClientSQL.client_name.in_(client_names))

    # order by 'created_utc' desc, so we get the latest one
    query = query.order_by(
        SiteSQL.site_uuid,
        DatetimeIntervalSQL.start_utc,
    )

    # get all results
    generations: List[GenerationSQL] = query.all()

    return generations


def get_pv_generation_by_sites(
    session: Session,
    start_utc: Optional[datetime] = None,
    end_utc: Optional[datetime] = None,
    site_uuids: Optional[List[uuid.UUID]] = None,
) -> List[GenerationSQL]:
    """
    Get the generation data by site

    :param session: database session
    :param start_utc: search filters >= on 'datetime_utc'
    :param end_utc: search fileters < on 'datetime_utc'
    :param site_uuids: optional list of site uuids
    :return: list of pv yields
    """

    # start main query
    query = session.query(GenerationSQL)
    query = query.join(SiteSQL)
    query = query.join(DatetimeIntervalSQL)

    # Filter by time interval
    query = filter_query_by_datetime_interval(query=query, start_utc=start_utc, end_utc=end_utc)

    if site_uuids is not None:
        query = query.filter(SiteSQL.site_uuid.in_(site_uuids))

    # Order by 'created_utc' desc
    query = query.order_by(SiteSQL.site_uuid, DatetimeIntervalSQL.start_utc)

    # get all results
    generations: List[GenerationSQL] = query.all()

    return generations



""" Pytest fixtures for tests """
from datetime import datetime, timedelta, timezone

import pytest
import uuid

from pvsite_datamodel.sqlmodels import Base
from pvsite_datamodel import (
    ClientSQL,
    SiteSQL,
    GenerationSQL,
    StatusSQL,
    LatestForecastValueSQL,
    ForecastSQL,
)
from pvsite_datamodel.write.datetime_intervals import get_or_else_create_datetime_interval

from testcontainers.postgres import PostgresContainer
from sqlalchemy import create_engine
from sqlalchemy.orm import Session

from typing import List


@pytest.fixture(scope="session")
def engine():
    with PostgresContainer("postgres:14.5") as postgres:
        # TODO need to setup postgres database with docker
        url = postgres.get_connection_url()
        engine = create_engine(url)
        Base.metadata.create_all(engine)

        yield engine


@pytest.fixture(scope="function", autouse=True)
def db_session(engine):
    """Returns an sqlalchemy session, and after the test tears down everything properly."""
    connection = engine.connect()
    # begin the nested transaction
    transaction = connection.begin()
    # use the connection with the already started transaction

    with Session(bind=connection) as session:
        yield session

        session.close()
        # roll back the broader transaction
        transaction.rollback()
        # put back the connection to the connection pool
        connection.close()
        session.flush()

    engine.dispose()


@pytest.fixture()
def sites(db_session):
    """create some fake sites"""

    sites = []
    for i in range(0, 4):
        client = ClientSQL(
            client_uuid=uuid.uuid4(),
            client_name=f"testclient_{i}",
            created_utc=datetime.now(timezone.utc),
        )
        site = SiteSQL(
            site_uuid=uuid.uuid4(),
            client_uuid=client.client_uuid,
            client_site_id=1,
            latitude=51,
            longitude=3,
            capacity_kw=4,
            created_utc=datetime.now(timezone.utc),
            updated_utc=datetime.now(timezone.utc),
            ml_id=i,
        )
        db_session.add(client)
        db_session.add(site)
        db_session.commit()

        sites.append(site)

    return sites


@pytest.fixture()
def generations(db_session, sites):
    """Create some fake generations"""

    start_times = [datetime.today() - timedelta(minutes=x) for x in range(10)]

    all_generations = []
    for site in sites:
        for i in range(0, 10):
            datetime_interval, _ = get_or_else_create_datetime_interval(
                session=db_session, start_time=start_times[i]
            )

            generation = GenerationSQL(
                generation_uuid=uuid.uuid4(),
                site_uuid=site.site_uuid,
                power_kw=i,
                datetime_interval_uuid=datetime_interval.datetime_interval_uuid,
            )
            all_generations.append(generation)

    db_session.add_all(all_generations)
    db_session.commit()


@pytest.fixture()
def latestforecastvalues(db_session, sites):
    """Create some fake latest forecast values"""

    latest_forecast_values = []
    forecast_version: str = "0.0.0"
    start_times = [datetime.today() - timedelta(minutes=x) for x in range(10)]

    for site in sites:
        forecast: ForecastSQL = ForecastSQL(
            forecast_uuid=uuid.uuid4(),
            site_uuid=site.site_uuid,
            forecast_version=forecast_version,
        )
        for i in range(0, 10):
            datetime_interval, _ = get_or_else_create_datetime_interval(
                session=db_session, start_time=start_times[i]
            )

            latest_forecast_value: LatestForecastValueSQL = LatestForecastValueSQL(
                latest_forecast_value_uuid=uuid.uuid4(),
                datetime_interval_uuid=datetime_interval.datetime_interval_uuid,
                forecast_generation_kw=i,
                forecast_uuid=forecast.forecast_uuid,
                site_uuid=site.site_uuid,
                forecast_version=forecast_version,
            )

            latest_forecast_values.append(latest_forecast_value)

    db_session.add_all(latest_forecast_values)
    db_session.commit()


@pytest.fixture()
def datetimeintervals(db_session):
    """Create fake datetime intervals"""

    start_times: List[datetime] = [datetime.today() - timedelta(minutes=x) for x in range(10)]

    for time in start_times:
        get_or_else_create_datetime_interval(session=db_session, start_time=time)


@pytest.fixture()
def test_time():
    return datetime(2022, 7, 25, 0, 0, 0, 0, timezone.utc)


@pytest.fixture()
def forecast_valid_site(sites):
    site_uuid = sites[0].site_uuid

    return {
        "target_datetime_utc": [
            (datetime.today() - timedelta(minutes=x)).isoformat() for x in range(10)
        ],
        "forecast_kw": [float(x) for x in range(10)],
        "pv_uuid": [site_uuid for x in range(10)],
    }


@pytest.fixture()
def forecast_invalid_site():
    return {
        "target_datetime_utc": [datetime.today().isoformat()],
        "forecast_kw": [1.0],
        "pv_uuid": [uuid.uuid4()],
    }


@pytest.fixture()
def generation_valid_site(sites):
    site_uuid = sites[0].site_uuid

    return {
        "start_datetime_utc": [(datetime.today() - timedelta(minutes=x)) for x in range(10)],
        "power_kw": [float(x) for x in range(10)],
        "site_uuid": [site_uuid for x in range(10)],
    }


@pytest.fixture()
def generation_invalid_site():
    return {
        "start_datetime_utc": [datetime.today()],
        "power_kw": [1.0],
        "site_uuid": [uuid.uuid4()],
    }


@pytest.fixture()
def statuses(db_session) -> List[StatusSQL]:
    """Create some fake statuses"""

    statuses: List[StatusSQL] = []
    for i in range(0, 4):
        status = StatusSQL(
            status_uuid=uuid.uuid4(),
            status="OK",
            message=f"Status {i}",
        )
        db_session.add(status)
        db_session.commit()
        statuses.append(status)

    return statuses



"""
Functions for writing to pvsite db
"""
import datetime as dt
import logging
import uuid

import numpy.typing as npt
import pandas as pd
import sqlalchemy.orm as sa_orm
from pvsite_datamodel.read.site import get_site_by_uuid
from pvsite_datamodel.sqlmodels import GenerationSQL

# Defines the length of time over which a forecast is valid
from pvsite_datamodel.write.datetime_intervals import get_or_else_create_datetime_interval
from pvsite_datamodel.write.upsert import upsert
from pvsite_datamodel.write.utils import WrittenRow


def insert_generation_values(
    session: sa_orm.Session,
    generation_values_df: pd.DataFrame,
) -> list[WrittenRow]:
    """
    Inserts a dataframe of forecast values into the database.

    :param session: sqlalchemy session for interacting with the database
    :param generation_values_df: pandas dataframe with columns
    ["start_datetime_utc", "power_kw", "pv_uuid"]
    :return list[WrittenRow]: list of added rows to DB
    """

    # Track rows added to DB
    written_rows: list[WrittenRow] = []

    # Loop over all the unique sites that have got forecast values
    site_uuids: npt.ndarray[uuid.UUID] = generation_values_df["site_uuid"].unique()
    generation_sqls = []
    for site_uuid in site_uuids:

        # Check whether the site id exits in the table, otherwise return an error
        get_site_by_uuid(session=session, site_uuid=site_uuid)

        # Get all dataframe forecast value entries for current site_uuid
        df_site: pd.DataFrame = generation_values_df.loc[
            generation_values_df["site_uuid"] == site_uuid
        ]

        # Filter the forecasted values by target_time
        start_datetimes: npt.ndarray[dt.datetime] = df_site["start_datetime_utc"].unique()

        # Print a warning if there are duplicate target_times for this site's forecast
        if len(start_datetimes) != len(df_site):
            logging.warning(f"duplicate target datetimes " f"for site {site_uuid}")

        # For each target time:
        for start_datetime in start_datetimes:
            datetime_interval, newly_added_rows = get_or_else_create_datetime_interval(
                session=session, start_time=pd.to_datetime(start_datetime)
            )
            written_rows.extend(newly_added_rows)

            # For each entry with this target time:
            df_target_entries: pd.DataFrame = df_site.loc[
                df_site["start_datetime_utc"] == start_datetimes
            ]

            # Create a GenerationSQL object for each generation, and surface as dict
            generation = GenerationSQL(
                site_uuid=site_uuid,
                generation_uuid=uuid.uuid4(),
                power_kw=df_target_entries.iloc[0]["power_kw"],
                datetime_interval_uuid=datetime_interval.datetime_interval_uuid,
            ).__dict__
            generation_sqls.append(generation)

        # Save it to the db
        newly_added_rows = upsert(session, GenerationSQL, generation_sqls)
        written_rows.extend(newly_added_rows)

    return written_rows



"""
Cross-package functions
"""

import datetime as dt
import typing

from pvsite_datamodel.sqlmodels import Base


class WrittenRow(typing.NamedTuple):
    """
    Defines a write to the Database
    """

    table: Base
    pk_value: str


FORECAST_TIMESPAN = dt.timedelta(minutes=5)



""" Test read functions """

import datetime as dt
import uuid

from sqlalchemy.orm import Query
from typing import List

from pvsite_datamodel import SiteSQL, StatusSQL, ClientSQL, DatetimeIntervalSQL
from pvsite_datamodel.read import get_all_sites
from pvsite_datamodel.read import get_site_by_uuid
from pvsite_datamodel.read import get_site_by_client_site_id
from pvsite_datamodel.read import get_pv_generation_by_sites
from pvsite_datamodel.read import get_latest_status
from pvsite_datamodel.read import get_pv_generation_by_client
from pvsite_datamodel.read import get_latest_forecast_values_by_site
from pvsite_datamodel.read.utils import filter_query_by_datetime_interval

import pytest


class TestGetAllSites:
    """Tests for the get_all_sites function"""

    def test_returns_all_sites(self, sites, db_session):
        out = get_all_sites(db_session)

        assert len(out) == len(sites)
        assert out[0] == sites[0]


class TestGetSiteByUUID:
    """Tests for the get_site_by_uuid function"""

    def tests_gets_site_for_existing_uuid(self, sites, db_session):
        site = get_site_by_uuid(session=db_session, site_uuid=sites[0].site_uuid)

        assert site == sites[0]

    def test_raises_error_for_nonexistant_site(self, sites, db_session):
        with pytest.raises(Exception):
            _ = get_site_by_uuid(session=db_session, site_uuid=uuid.uuid4())


class TestGetSiteByClientSiteID:
    """Tests for the get_site_by_client_site_id function"""

    def test_gets_site_successfully(self, sites, db_session):
        site = get_site_by_client_site_id(
            session=db_session,
            client_name="testclient_1",
            client_site_id=1
        )

        assert site.client_site_id == 1

    def test_raises_exception_when_no_such_site_exists(self, sites, db_session):
        with pytest.raises(Exception):
            _ = get_site_by_client_site_id(
                session=db_session,
                client_name="testclient_100",
                client_site_id=1
            )


class TestGetPVGenerationByClient:
    """Tests for the get_pv_generation_by_client function"""

    def test_returns_all_generations_without_input_client(self, generations, db_session):
        generations = get_pv_generation_by_client(session=db_session)

        assert len(generations) == 40

    def test_returns_all_generations_for_input_client(self, generations, db_session):
        query: Query = db_session.query(ClientSQL)
        client: ClientSQL = query.first()

        generations = get_pv_generation_by_client(
            session=db_session,
            client_names=[client.client_name]
        )

        assert len(generations) == 10

    def test_returns_all_generations_in_datetime_window(self, generations, db_session):
        query: Query = db_session.query(ClientSQL)
        client: ClientSQL = query.first()
        window_lower: dt.datetime = dt.datetime.today() - dt.timedelta(minutes=7)
        window_upper: dt.datetime = dt.datetime.today() + dt.timedelta(minutes=8)

        generations = get_pv_generation_by_client(
            session=db_session,
            client_names=[client.client_name],
            start_utc=window_lower,
            end_utc=window_upper
        )

        assert len(generations) == 7


class TestGetPVGenerationBySites:
    """Tests for the get_pv_generation_by_sites function"""

    def test_gets_generation_for_single_input_site(self, generations, db_session):
        query: Query = db_session.query(SiteSQL)
        site: SiteSQL = query.first()

        generations = get_pv_generation_by_sites(
            session=db_session,
            site_uuids=[site.site_uuid]
        )

        assert len(generations) == 10
        assert generations[0].datetime_interval is not None

    def test_gets_generation_for_multiple_input_sites(self, generations, db_session):
        query: Query = db_session.query(SiteSQL)
        sites: List[SiteSQL] = query.all()

        generations = get_pv_generation_by_sites(
            session=db_session,
            site_uuids=[site.site_uuid for site in sites]
        )

        assert len(generations) == 10 * len(sites)

    def test_returns_empty_list_for_no_input_sites(self, generations, db_session):
        generations = get_pv_generation_by_sites(
            session=db_session,
            site_uuids=[]
        )

        assert len(generations) == 0


class TestGetLatestStatus:
    """Tests for the get_latest_status function"""

    def test_gets_latest_status_when_exists(self, statuses, db_session):
        status: StatusSQL = get_latest_status(db_session)

        assert status.message == "Status 3"


class TestGetLatestForecastValuesBySite:
    """Tests for the get_latest_forecast_values_by_site function"""

    def test_gets_latest_forecast_values_with_single_site(self, latestforecastvalues, db_session):
        query: Query = db_session.query(SiteSQL)
        site: SiteSQL = query.first()

        latest_forecast_values = get_latest_forecast_values_by_site(
            session=db_session,
            site_uuids=[site.site_uuid]
        )

        assert len(latest_forecast_values) == 1
        assert len(latest_forecast_values[site.site_uuid]) == 10
        assert latest_forecast_values[site.site_uuid][0].datetime_interval is not None

    def test_gets_latest_forecast_values_with_multiple_sites(
            self, latestforecastvalues, db_session):
        query: Query = db_session.query(SiteSQL)
        sites: SiteSQL = query.all()

        latest_forecast_values = get_latest_forecast_values_by_site(
            session=db_session,
            site_uuids=[site.site_uuid for site in sites]
        )

        assert len(latest_forecast_values) == len(sites)

    def test_gets_latest_forecast_values_filter_start_utc(
            self, latestforecastvalues, db_session):
        query: Query = db_session.query(SiteSQL)
        site: SiteSQL = query.first()

        latest_forecast_values = get_latest_forecast_values_by_site(
            session=db_session,
            site_uuids=[site.site_uuid],
            start_utc=dt.datetime.today() - dt.timedelta(minutes=7)
        )
        assert len(latest_forecast_values[site.site_uuid]) == 7

        latest_forecast_values = get_latest_forecast_values_by_site(
            session=db_session,
            site_uuids=[site.site_uuid],
            start_utc=dt.datetime.today() - dt.timedelta(minutes=5)
        )
        assert len(latest_forecast_values[site.site_uuid]) == 5


class TestFilterQueryByDatetimeInterval:
    """Tests for the filter_query_by_datetime_interval function"""

    def test_returns_datetime_intervals_in_filter(self, datetimeintervals, db_session):
        query: Query = db_session.query(DatetimeIntervalSQL)
        query = filter_query_by_datetime_interval(
            query=query,
            start_utc=dt.datetime.today() - dt.timedelta(minutes=7)
        )

        datetime_intervals: List[DatetimeIntervalSQL] = query.all()

        assert len(datetime_intervals) == 7



"""
Functions for reading to pvsite db
"""

import logging
from typing import List, Optional

from pvsite_datamodel.sqlmodels import ClientSQL, SiteSQL
from sqlalchemy.orm import Session

logger = logging.getLogger(__name__)


def get_site_by_uuid(session: Session, site_uuid: str) -> SiteSQL:
    """
    Get site object from uuid.

    Raise error if site does not exist

    :param session: database sessions
    :param site_uuid: the site uuid
    :return: the site object
    """
    query = session.query(SiteSQL)
    query = query.filter(SiteSQL.site_uuid == site_uuid)
    existing_site: Optional[SiteSQL] = query.first()
    if existing_site is None:
        raise KeyError(f"Site uuid {site_uuid} not found in sites table")

    return existing_site


def get_site_by_client_site_id(session: Session, client_name: str, client_site_id: int) -> SiteSQL:
    """
    Get site from client name and client site id

    :param session: database sessions
    :param client_name: client name
    :param client_site_id: client's id of site
    :return: site object, or None
    """

    logger.debug(f"Getting {client_name}'s site {client_site_id}")

    # start main query
    query = session.query(SiteSQL)
    query = query.join(ClientSQL)

    # select the correct client site id
    query = query.filter(SiteSQL.client_site_id == client_site_id)

    # filter on client_name
    query = query.filter(ClientSQL.client_name == client_name)

    # get first result (should only be one site)
    site: Optional[SiteSQL] = query.first()

    if site is None:
        raise Exception(f"Could not find site {client_site_id} from client {client_name}")

    return site


def get_site_by_client_site_name(
    session: Session, client_name: str, client_site_name: str
) -> SiteSQL:
    """
    Get site from client name and client site id

    :param session: database sessions
    :param client_name: client name
    :param client_site_name: client's name of site
    :return: site object, or None
    """

    logger.debug(f"Getting {client_name}'s site {client_site_name}")

    # start main query
    query = session.query(SiteSQL)
    query = query.join(ClientSQL)

    # select the correct client site name
    query = query.filter(SiteSQL.client_site_name == client_site_name)

    # filter on client_name
    query = query.filter(ClientSQL.client_name == client_name)

    # get first result (should only be one site)
    site: Optional[SiteSQL] = query.first()

    if site is None:
        raise Exception(f"Could not find site {client_site_name} from client {client_name}")

    return site


def get_all_sites(session: Session) -> List[SiteSQL]:
    """
    Gets all sites from sites table

    :param session: database sessions
    :return: site object
    """

    logger.debug("Getting all sites")

    # start main query
    query = session.query(SiteSQL)

    # get all results
    sites = query.all()

    logger.debug(f"Found {len(sites)} sites")

    return sites
